from typing import Callable, Optional, Any

import tensorflow as tf
from tensorflow import Tensor, DType
from tensorflow.keras.metrics import Metric


class Mean(Metric):
    def __init__(
            self,
            function: Callable[[Any, Any, Any], Tensor],
            name: str,
            dtype: DType = tf.float32,
    ) -> None:

        super().__init__(name=name, dtype=dtype)

        self.function = function

        self.mean = self.add_weight('mean', initializer=tf.zeros_initializer, dtype=dtype)
        self.count = self.add_weight('count', initializer=tf.zeros_initializer, dtype=dtype)

    def update_mean(self, values: Tensor, batch_size: Tensor) -> None:
        new_count = self.count + batch_size
        new_mean = (tf.reduce_sum(values) + self.count * self.mean) / new_count

        self.mean.assign(new_mean)

    def update_state(self, y_true, y_pred, sample_weight: Optional[Tensor] = None) -> None:
        values = self.function(y_true, y_pred, sample_weight)
        batch_size = tf.cast(tf.size(values), self.dtype)

        self.update_mean(values, batch_size)
        self.count.assign_add(batch_size)

    def result(self) -> Tensor:
        return self.mean

    def reset_state(self) -> None:
        self.mean.assign(1, tf.zeros((), self.dtype))
        self.count.assign(tf.zeros((), self.dtype))



from typing import List, Optional, Mapping, Any

import tensorflow as tf
from tensorflow.keras.layers import Layer, Dense

from sfu_ml_lib.common import math
from sfu_ml_lib.common import shaping


class MultiHeadDense(Layer):
    def __init__(
            self,
            units: int,
            num_heads: int,
            activation=None,
            use_bias: bool = True,
            kernel_initializer='glorot_uniform',
            bias_initializer='zeros',
            kernel_regularizer=None,
            bias_regularizer=None,
            activity_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
            **kwargs) -> None:

        super().__init__(activity_regularizer=tf.keras.regularizers.get(activity_regularizer), **kwargs)

        self.units = units
        self.num_heads = num_heads
        self.use_bias = use_bias

        self.activation = tf.keras.activations.get(activation)
        self.kernel_initializer = tf.initializers.get(kernel_initializer)
        self.bias_initializer = tf.initializers.get(bias_initializer)
        self.kernel_regularizer = tf.keras.regularizers.get(kernel_regularizer)
        self.bias_regularizer = tf.keras.regularizers.get(bias_regularizer)
        self.kernel_constraint = tf.keras.constraints.get(kernel_constraint)
        self.bias_constraint = tf.keras.constraints.get(bias_constraint)

        self.kernel: Optional[Layer] = None
        self.bias: Optional[Layer] = None

    def build(self, input_shape: List[int]) -> None:
        self.kernel = self.add_weight(
            name='kernel',
            shape=[input_shape[-2], self.units, self.num_heads],
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            dtype=self.dtype,
            trainable=True,
        )

        if self.use_bias:
            self.bias = self.add_weight(
                name='bias',
                shape=[self.units, self.num_heads],
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                dtype=self.dtype,
                trainable=True,
            )

        self.built = True

    def call(self, inputs: tf.Tensor, **kwargs) -> tf.Tensor:
        outputs = tf.einsum('...ih,ijh->...jh', inputs, self.kernel)

        if self.use_bias:
            outputs = tf.add(outputs, self.bias)

        if self.activation is not None:
            outputs = self.activation(outputs)

        return outputs

    def get_config(self) -> Mapping[str, Any]:
        new_config = {
            'units': self.units,
            'use_bias': self.use_bias,
            'activation': tf.keras.activations.serialize(self.activation),
            'kernel_initializer': tf.initializers.serialize(self.kernel_initializer),
            'bias_initializer': tf.initializers.serialize(self.bias_initializer),
            'kernel_regularizer': tf.keras.regularizers.serialize(self.kernel_regularizer),
            'bias_regularizer': tf.keras.regularizers.serialize(self.bias_regularizer),
            'activity_regularizer': tf.keras.regularizers.serialize(self.activity_regularizer),
            'kernel_constraint': tf.keras.constraints.serialize(self.kernel_constraint),
            'bias_constraint': tf.keras.constraints.serialize(self.bias_constraint)
        }

        config = super().get_config()
        config.update(new_config)

        return config


class MultiHeadAttention(Layer):
    def __init__(
            self,
            num_heads: int,
            temperature: float = 1.,
            key_transformer: Optional[Layer] = None,
            value_transformer: Optional[Layer] = None,
            name: Optional[str] = None) -> None:

        super().__init__(name=name)

        self.num_heads = num_heads
        self.temperature = temperature
        self.key_transformer = key_transformer
        self.value_transformer = value_transformer

        self.coefficient_kernel: Optional[tf.Variable] = None
        self.coefficient_bias: Optional[tf.Variable] = None
        self.dense: Optional[Layer] = None

    def build(self, input_shape: List[int]) -> None:
        attention_size = input_shape[-1]

        self.coefficient_kernel = self.add_weight(
            name='coefficient_kernel',
            shape=(attention_size, self.num_heads),
            initializer=tf.initializers.glorot_uniform(),
            dtype=tf.float32,
            trainable=True,
        )

        self.coefficient_bias = self.add_weight(
            name='coefficient_bias',
            shape=(self.num_heads,),
            initializer=tf.initializers.zeros(),
            dtype=tf.float32,
            trainable=True,
        )

        self.dense = Dense(attention_size)

        self.built = True

    def compute_mask(self, inputs: tf.Tensor, mask: Optional[tf.Tensor] = None) -> Optional[tf.Tensor]:
        if mask is None:
            return None

        mask = tf.reduce_any(mask, axis=-1)

        return mask

    def call(self, inputs: tf.Tensor, mask: Optional[tf.Tensor] = None, training: bool = False) -> tf.Tensor:
        assert self.dense

        if mask is None:
            mask = tf.ones(tf.shape(inputs)[:-1], dtype=tf.bool)

        coefficients = values = shaping.tile_dimension(inputs, self.num_heads)

        if self.key_transformer:
            coefficients = self.key_transformer(coefficients, training=training)

        coefficients = tf.einsum('...ih,ih->...h', coefficients, self.coefficient_kernel)
        coefficients = tf.nn.bias_add(coefficients, self.coefficient_bias)
        coefficients /= self.temperature
        coefficients = math.softmax_masked(coefficients, mask)

        if self.value_transformer:
            values = self.value_transformer(values, training=training)

        values = tf.einsum('...sh,...sih->...ih', coefficients, values)
        values = shaping.flatten(values)
        values = self.dense(values)

        return values


class MultiHeadSelfAttention(Layer):
    def __init__(self, num_heads: int) -> None:
        super().__init__()

        self.num_heads = num_heads

        self.temperature = None

        self.query_layer: Optional[Layer] = None
        self.key_layer: Optional[Layer] = None
        self.value_layer: Optional[Layer] = None
        self.dense: Optional[Layer] = None

    def build(self, input_shape: List[int]) -> None:
        attention_size = input_shape[-1]

        self.temperature = tf.sqrt(tf.cast(attention_size, tf.float32))

        self.query_layer = MultiHeadDense(attention_size, self.num_heads)
        self.key_layer = MultiHeadDense(attention_size, self.num_heads)
        self.value_layer = MultiHeadDense(attention_size, self.num_heads)
        self.dense = Dense(attention_size)

        self.built = True

    def call(self, inputs: tf.Tensor, mask: Optional[tf.Tensor] = None) -> tf.Tensor:
        assert self.query_layer and self.key_layer and self.value_layer and self.dense

        if mask is None:
            mask = tf.ones(tf.shape(inputs)[:-1], dtype=tf.bool)

        inputs = shaping.tile_dimension(inputs, self.num_heads)

        queries = self.query_layer(inputs)
        keys = self.key_layer(inputs)

        coefficients = tf.einsum('...qih,...kih->...qkh', queries, keys)
        coefficients /= self.temperature
        coefficients = math.softmax_masked(coefficients, mask)

        values = self.value_layer(inputs)
        values = tf.einsum('...qkh,...qih->...qih', coefficients, values)
        values = shaping.flatten(values)
        values = self.dense(values)

        return values



import typing
from abc import ABC, abstractmethod
from typing import Tuple, Optional, Iterator, TypeVar, Union

import numpy as np
import tensorflow as tf
from tensorflow import Tensor


T = TypeVar('T')
TensorType = Union[np.ndarray, Tensor]
TensorTuple = Union[TensorType, Tuple[TensorType, ...]]
BatchType = Union[
    Tuple[TensorTuple, None],
    Tuple[TensorTuple, TensorTuple],
    Tuple[TensorTuple, TensorTuple, TensorType],
]
DatasetType = Union[Iterator[BatchType], tf.data.Dataset]


class Dataset(ABC):
    size: int
    targets: Optional[TensorType] = None

    @abstractmethod
    @typing.no_type_check
    def get_batches(self, *args) -> DatasetType:
        ...

    @abstractmethod
    def subset(self: T, indices: TensorType) -> T:
        ...



import tensorflow as tf


def segment_softmax(logits: tf.Tensor, segment_ids: tf.Tensor) -> tf.Tensor:
    logits_max = tf.math.segment_max(logits, segment_ids)

    logits_exp = tf.math.exp(logits - tf.gather(logits_max, segment_ids))

    partitions = tf.gather(tf.math.segment_sum(logits_exp, segment_ids), segment_ids)

    softmax = logits_exp / partitions

    return softmax


def segment_normalize(logits: tf.Tensor, segment_ids: tf.Tensor) -> tf.Tensor:
    partitions = tf.gather(tf.math.segment_sum(logits, segment_ids), segment_ids)

    probabilities = logits / partitions

    return probabilities



from typing import List, Optional

import tensorflow as tf
from tensorflow.keras.layers import Dense, LayerNormalization, Dropout
from tensorflow.keras.layers import Layer
from tensorflow.keras.models import Sequential

from sfu_ml_lib.modules.attention import MultiHeadSelfAttention


class SelfAttentionEncoderLayer(Layer):
    def __init__(self, num_heads: int, dropout_rate: float) -> None:
        super().__init__()

        self.num_heads = num_heads
        self.dropout_rate = dropout_rate

        self.dropout = Dropout(dropout_rate)
        self.layer_normalization_first = LayerNormalization()
        self.layer_normalization_second = LayerNormalization()

        self.attention: Optional[Layer] = None
        self.feed_forward: Optional[Layer] = None

        self.supports_masking = True

    def build(self, input_shape: List[int]) -> None:
        attention_size = input_shape[-1]

        self.attention = MultiHeadSelfAttention(self.num_heads)

        self.feed_forward = Sequential([
            Dense(attention_size, activation=tf.nn.relu),
            Dropout(self.dropout_rate),
            Dense(attention_size),
        ])

        self.built = True

    def call(self, inputs: tf.Tensor, training: bool = False, mask: Optional[tf.Tensor] = None) -> tf.Tensor:
        assert self.attention and self.feed_forward

        attention_output = self.attention(inputs, mask)
        attention_output = self.dropout(attention_output, training=training)
        attention_output = self.layer_normalization_first(inputs + attention_output)

        output = self.feed_forward(attention_output)
        output = self.dropout(output, training=training)
        output = self.layer_normalization_second(attention_output + output)

        return output



import inspect
import sys
from typing import Callable, Union, Optional, List, Mapping, Set


PrimitiveType = Union[int, float, bool, str]


class ScriptParameters:
    @classmethod
    def get_positional_arguments(cls, argv: Optional[List[str]] = None) -> List[str]:
        argv = argv if argv else sys.argv[1:]

        arguments = []

        for argument in argv:
            if argument.startswith('-'):
                break

            arguments.append(argument)

        return arguments

    @classmethod
    def get_keyword_arguments(cls, argv: Optional[List[str]] = None) -> Set[str]:
        argv = argv if argv else sys.argv[1:]

        num_positional_arguments = len(cls.get_positional_arguments(argv))

        passed_arguments = {
            key.lstrip('-').replace('_', '-')
            for key in argv[num_positional_arguments::2]
            if key.startswith('-')
        }

        return passed_arguments

    @classmethod
    def get_script_parameters(
            cls,
            function: Callable,
            ignore_keyword_arguments: bool = True,
    ) -> Mapping[str, PrimitiveType]:
        """
        Returns the arguments of a function with its values specified by the command line or its default values.
        Underscores in the name of the arguments are transformed to dashes.
        Can optionally filter out keyword arguments obtained through the command line.
        :param function: the function to inspect.
        :param ignore_keyword_arguments: whether to filter out keyword command line arguments.
        :return: a map from argument names to default values.
        """
        positional_arguments, keyword_arguments = cls.get_positional_arguments(), cls.get_keyword_arguments()
        signature = inspect.signature(function)

        arguments = {}

        for index, (name, parameter) in enumerate(signature.parameters.items()):
            transformed_name = name.replace('_', '-')

            if index < len(positional_arguments):
                arguments[transformed_name] = positional_arguments[index]

            elif not (ignore_keyword_arguments and transformed_name in keyword_arguments):
                if parameter.default != parameter.empty:
                    arguments[transformed_name] = parameter.default

        return arguments



import math
from typing import Tuple, Generator, Optional, List, Union, TypeVar

import numpy as np
from sklearn.model_selection import StratifiedShuffleSplit, StratifiedKFold, KFold, ShuffleSplit

from sfu_ml_lib.base.dataset import Dataset


SizeType = Union[int, float]
DatasetType = TypeVar('DatasetType', bound=Dataset)


def get_dataset_folds(
        dataset: DatasetType,
        number_folds: int = 10,
        validation_size: Optional[SizeType] = None,
        random_state: int = 110069) -> Generator[Tuple[DatasetType, Optional[DatasetType], DatasetType], None, None]:
    """
    Creates multiple folds suitable for cross-validation, meaning there's no overlap of example within the test sets
    produced in each fold.
    :param dataset: A `Dataset` object.
    :param number_folds: Number of folds to use.
    :param validation_size: If float this is the proportion of the training split to use as validation. Having it
    set to None or zero means there is no validation set.
    :param random_state: A seed for random number generation.
    :return: A generator of size `number_folds` with a tuple of up to three datasets corresponding to the splits.
    """
    train_test_folder = KFold(n_splits=number_folds, shuffle=True, random_state=random_state)

    for train_validation_indices, test_indices in train_test_folder.split(np.empty(dataset.size, dtype=np.bool)):
        test_dataset = dataset.subset(test_indices)

        if not validation_size:
            train_dataset = dataset.subset(train_validation_indices)

            yield train_dataset, None, test_dataset

        else:
            train_validation_splitter = ShuffleSplit(n_splits=1, test_size=validation_size, random_state=random_state)

            train_validation_inputs = np.empty(train_validation_indices.size, dtype=np.bool)
            subset_train_indices, subset_validation_indices = next(train_validation_splitter.split(
                X=train_validation_inputs,
            ))

            train_dataset = dataset.subset(train_validation_indices[subset_train_indices])
            validation_dataset = dataset.subset(train_validation_indices[subset_validation_indices])

            yield train_dataset, validation_dataset, test_dataset


def get_dataset_splits(
        dataset: DatasetType,
        train_size: SizeType,
        validation_size: Optional[SizeType] = None,
        test_size: Optional[SizeType] = None,
        number_splits: int = 10,
        random_state: int = 110069) -> Generator[
            Tuple[DatasetType, Optional[DatasetType], Optional[DatasetType]],
            None,
            None]:
    """
    Creates multiple splits of the dataset.
    :param dataset: A `Dataset` object.
    :param train_size: Number of examples or a proportion of the dataset.
    :param validation_size: Number of examples or a proportion of the dataset. Having it set to None means it uses the
    complement of the train set and the test set. If both `validation_size` and `test_size` are None then
    a validation set is created with the complement of the train set.
    :param test_size: Number of examples or a proportion of the dataset. Having it set to None means it uses the
    complement of the train set and the validation set. If both `validation_size` and `test_size` are None then
    a test set is not created.
    :param number_splits: Number of splits to generate.
    :param random_state: A seed for random number generation.
    :return: A generator of size `number_splits` with a tuple of up to three datasets corresponding to the splits.
    """
    train_size, validation_size, test_size = get_sizes(dataset.size, train_size, validation_size, test_size)

    train_test_splitter = ShuffleSplit(
        n_splits=number_splits,
        train_size=train_size,
        test_size=validation_size + test_size,
        random_state=random_state,
    )

    for train_indices, validation_test_indices in train_test_splitter.split(np.empty(dataset.size, dtype=np.bool)):
        train_dataset = dataset.subset(train_indices)

        if not validation_size and not test_size:
            yield train_dataset, None, None

        if not test_size:
            validation_dataset = dataset.subset(validation_test_indices)

            yield train_dataset, validation_dataset, None

        else:
            validation_test_splitter = ShuffleSplit(
                n_splits=1,
                train_size=validation_size,
                test_size=test_size,
                random_state=random_state,
            )

            validation_test_inputs = np.empty(validation_test_indices.size, dtype=np.bool)
            subset_validation_indices, subset_test_indices = next(validation_test_splitter.split(
                X=validation_test_inputs,
            ))

            validation_dataset = dataset.subset(validation_test_indices[subset_validation_indices])
            test_dataset = dataset.subset(validation_test_indices[subset_test_indices])

            yield train_dataset, validation_dataset, test_dataset


def get_stratified_dataset_folds(
        dataset: DatasetType,
        number_folds: int = 10,
        validation_size: Optional[SizeType] = None,
        random_state: int = 110069) -> Generator[Tuple[DatasetType, Optional[DatasetType], DatasetType], None, None]:
    """
    Creates multiple folds suitable for cross-validation, meaning there's no overlap of example within the test sets
    produced in each fold.
    Tries to maintain the same distribution of classes in the dataset across types and folds.
    :param dataset: A `Dataset` object.
    :param number_folds: Number of folds to use.
    :param validation_size: If float this is the proportion of the training split to use as validation. Having it
    set to None or zero means there is no validation set.
    :param random_state: A seed for random number generation.
    :return: A generator of size `number_folds` with a tuple of up to three datasets corresponding to the splits.
    """
    assert dataset.targets is not None

    train_test_folder = StratifiedKFold(n_splits=number_folds, shuffle=True, random_state=random_state)

    fake_data = np.zeros_like(dataset.targets)

    for train_validation_indices, test_indices in train_test_folder.split(fake_data, dataset.targets):
        test_dataset = dataset.subset(test_indices)

        if not validation_size:
            train_dataset = dataset.subset(train_validation_indices)

            yield train_dataset, None, test_dataset

        else:
            train_validation_splitter = StratifiedShuffleSplit(
                n_splits=1,
                test_size=validation_size,
                random_state=random_state,
            )

            train_validation_targets = dataset.targets[train_validation_indices]

            subset_train_indices, subset_validation_indices = next(train_validation_splitter.split(
                X=np.zeros_like(train_validation_targets),
                y=train_validation_targets,
            ))

            train_indices = train_validation_indices[subset_train_indices]
            validation_indices = train_validation_indices[subset_validation_indices]

            train_dataset = dataset.subset(train_indices)
            validation_dataset = dataset.subset(validation_indices)

            yield train_dataset, validation_dataset, test_dataset


def get_stratified_dataset_splits(
        dataset: DatasetType,
        train_size: SizeType,
        validation_size: Optional[SizeType] = None,
        test_size: Optional[SizeType] = None,
        number_splits: int = 10,
        random_state: int = 110069) -> Generator[
            Tuple[DatasetType, Optional[DatasetType], Optional[DatasetType]],
            None,
            None]:
    """
    Creates multiple splits of the dataset.
    Tries to maintain the same distribution of classes in the dataset across types and splits.
    :param dataset: A `Dataset` object.
    :param train_size: Number of examples or a proportion of the dataset.
    :param validation_size: Number of examples or a proportion of the dataset. Having it set to None means it uses the
    complement of the train set and the test set. If both `validation_size` and `test_size` are None then
    a validation set is created with the complement of the train set.
    :param test_size: Number of examples or a proportion of the dataset. Having it set to None means it uses the
    complement of the train set and the validation set. If both `validation_size` and `test_size` are None then
    a test set is not created.
    :param number_splits: Number of splits to generate.
    :param random_state: A seed for random number generation.
    :return: A generator of size `number_splits` with a tuple of up to three datasets corresponding to the splits.
    """
    assert dataset.targets is not None

    train_size, validation_size, test_size = get_sizes(dataset.size, train_size, validation_size, test_size)

    train_test_splitter = StratifiedShuffleSplit(
        n_splits=number_splits,
        train_size=train_size,
        test_size=validation_size + test_size,
        random_state=random_state,
    )

    fake_data = np.zeros_like(dataset.targets)

    for train_indices, validation_test_indices in train_test_splitter.split(fake_data, dataset.targets):
        train_dataset = dataset.subset(train_indices)

        if not validation_size and not test_size:
            yield train_dataset, None, None

        if not test_size:
            validation_dataset = dataset.subset(validation_test_indices)

            yield train_dataset, validation_dataset, None

        else:
            validation_test_splitter = StratifiedShuffleSplit(
                n_splits=1,
                train_size=validation_size,
                test_size=test_size,
                random_state=random_state,
            )

            validation_test_targets = dataset.targets[validation_test_indices]

            subset_validation_indices, subset_test_indices = next(validation_test_splitter.split(
                X=np.zeros_like(validation_test_targets),
                y=validation_test_targets,
            ))

            validation_indices = validation_test_indices[subset_validation_indices]
            test_indices = validation_test_indices[subset_test_indices]

            validation_dataset = dataset.subset(validation_indices)
            test_dataset = dataset.subset(test_indices)

            yield train_dataset, validation_dataset, test_dataset


def get_sizes(
        dataset_size: int,
        train_size: SizeType,
        validation_size: Optional[SizeType],
        test_size: Optional[SizeType]) -> Tuple[int, int, int]:

    if isinstance(train_size, float):
        train_size = math.floor(dataset_size * train_size)

    if isinstance(validation_size, float):
        validation_size = math.floor(dataset_size * validation_size)

    if isinstance(test_size, float):
        test_size = math.floor(dataset_size * test_size)

    if validation_size is None and test_size is None:
        validation_size, test_size = dataset_size - train_size, 0

    elif validation_size and test_size is None:
        test_size = dataset_size - (train_size + validation_size)

    elif validation_size is None and test_size:
        validation_size = dataset_size - (train_size + test_size)

    validation_size = validation_size if validation_size else 0
    test_size = test_size if test_size else 0

    return train_size, validation_size, test_size


def train_test_split(
        data: List[np.ndarray],
        data_size: int,
        percentage: float) -> Tuple[List[np.ndarray], List[np.ndarray]]:

    threshold = int(data_size * percentage)

    return [i[:threshold] for i in data], [i[threshold:] for i in data]



import json
import os
from abc import ABC
from typing import List, Dict, Mapping, Tuple, Optional, TypeVar, Union, Callable, Iterable, Sequence, Any

import sfu_data_io.helpers as io
import tensorflow as tf
from tensorflow import Tensor, TensorSpec, RaggedTensorSpec, SparseTensorSpec
from tensorflow.keras import Model
from tensorflow.keras.callbacks import Callback
from tensorflow.keras.losses import Loss
from tensorflow.keras.metrics import Metric
from tensorflow.keras.optimizers import Optimizer

import sfu_ml_lib.base.metrics as metric_helpers
from sfu_ml_lib.base.dataset import DatasetType
from sfu_ml_lib.base.metrics import MetricKeyType, NumericType
from sfu_ml_lib.metrics.aggregator import KerasMetric


T = TypeVar('T')
W = TypeVar('W')
Z = TypeVar('Z')
BatchFunction = Callable[[], DatasetType]
Specification = Union[SparseTensorSpec, RaggedTensorSpec, TensorSpec]
Schema = Union[Specification, Iterable['Schema']]  # type: ignore


class KerasModel(Model, ABC):
    input_schema: Schema

    @classmethod
    def deconstruct_inputs(cls, features: T, targets: W, weights: Optional[Z] = None) -> Tuple[T, W, Optional[Z]]:
        """
        Converts the outputs of a tf.Dataset into features, targets and optional weights.
        Used in combination with `deconstruct_inputs(*args)`.
        """
        return features, targets, weights

    @classmethod
    def from_config_path(cls, path: str, **kwargs) -> 'KerasModel':
        with io.open(os.path.join(path, 'config.json'), mode='r') as json_file:
            config = json.load(json_file)

        config.update(kwargs)

        return cls.from_config(config)

    @classmethod
    def from_config(cls, config: Mapping[str, Any], custom_objects=None) -> 'KerasModel':
        return cls(**config)

    def save_config(self, path: str) -> None:
        with io.open(os.path.join(path, 'config.json'), mode='w') as json_file:
            json.dump(self.get_config(), json_file)

    @classmethod
    def convert_metrics(cls, metrics: Mapping[str, Tensor]) -> Mapping[MetricKeyType, NumericType]:
        return {label: value.numpy() for label, value in metrics.items()}

    @staticmethod
    def get_metrics(metrics: Sequence[Metric]) -> Mapping[str, Tensor]:
        metric_map: Dict[str, Tensor] = {}

        for metric in metrics:
            if isinstance(metric, KerasMetric):
                metric_map.update(metric.get_metrics())
            else:
                metric_map[metric.name] = metric.result()

        return metric_map

    def create_tf_dataset(self, batch_generator: BatchFunction, queue_size: int = -1) -> tf.data.Dataset:
        batches = batch_generator()

        if not isinstance(batches, tf.data.Dataset):
            batches = tf.data.Dataset.from_generator(batch_generator, output_signature=self.input_schema)

        return batches.prefetch(queue_size)

    def train_dataset(
            self,
            dataset: tf.data.Dataset,
            loss: Loss,
            optimizer: Optimizer,
            metrics: Sequence[Metric],
    ) -> Mapping[str, Tensor]:

        for metric in metrics:
            metric.reset_state()

        for arguments in dataset:
            features, targets, weights = self.deconstruct_inputs(*arguments)

            with tf.GradientTape() as tape:
                predictions = self(features, training=True)

                loss_result = loss(targets, predictions, weights)
                if len(self.losses) > 0:
                    loss_result += tf.add_n(self.losses)

            optimizer.minimize(loss_result, self.trainable_weights, tape=tape)

            for metric in metrics:
                metric.update_state(targets, predictions, weights)

        metric_results = self.get_metrics(metrics)

        return metric_results

    def test_dataset(self, dataset: tf.data.Dataset, metrics: Sequence[Metric]) -> Mapping[str, Tensor]:
        for metric in metrics:
            metric.reset_state()

        for arguments in dataset:
            features, targets, weights = self.deconstruct_inputs(*arguments)

            predictions = self(features, training=False)

            for metric in metrics:
                metric.update_state(targets, predictions, weights)

        metric_results = self.get_metrics(metrics)

        return metric_results

    def fit_dataset(
            self,
            train_batches: BatchFunction,
            max_num_epochs: int = 1,
            validation_batches: Optional[BatchFunction] = None,
            test_batches: Optional[BatchFunction] = None,
            callbacks: Optional[List[Callback]] = None,
    ) -> None:
        """
        Please look at `train_dataset` for details on how to construct `batches`.
        """
        callbacks = callbacks if callbacks else []
        metrics = self.compiled_metrics._metrics
        train_dataset = tf.function(self.train_dataset)
        test_dataset = tf.function(self.test_dataset)

        self.stop_training = False

        for callback in callbacks:
            callback.set_model(self)
            callback.on_train_begin()

        for epoch in range(max_num_epochs):
            for callback in callbacks:
                callback.on_epoch_begin(epoch)

            train_metrics = self.convert_metrics(train_dataset(
                dataset=self.create_tf_dataset(train_batches),
                loss=self.loss,
                optimizer=self.optimizer,
                metrics=metrics,
            ))
            metric_results = metric_helpers.prefix_metrics_train(train_metrics)

            if validation_batches is not None:
                validation_metrics = self.convert_metrics(test_dataset(
                    dataset=self.create_tf_dataset(validation_batches),
                    metrics=metrics,
                ))
                metric_results.update(metric_helpers.prefix_metrics_validation(validation_metrics))

            for callback in callbacks:
                callback.on_epoch_end(epoch, metric_results)

            if self.stop_training:
                break

        results = None

        if test_batches is not None:
            test_metrics = self.convert_metrics(test_dataset(
                dataset=self.create_tf_dataset(test_batches),
                metrics=metrics,
            ))
            results = metric_helpers.prefix_metrics_test(test_metrics)

        for callback in callbacks:
            callback.on_train_end(results)



from typing import Optional

import numpy as np
import sklearn.metrics as sk_metrics
from scipy import special as sp_scipy
from scipy import stats as sp_stats

from sfu_ml_lib.base.metrics import NumericType, AggregatorType


def logits_to_probabilities_multiclass(logits: np.ndarray) -> np.ndarray:
    return sp_scipy.softmax(logits, axis=-1)


def logits_to_probabilities_multilabel(logits: np.ndarray) -> np.ndarray:
    return sp_scipy.expit(logits, axis=-1)


def logits_to_predictions_multiclass(logits: np.ndarray) -> np.ndarray:
    return np.argmax(logits, axis=-1)


def logits_to_predictions_multilabel(logits: np.ndarray, dtype: np.dtype = np.float32) -> np.ndarray:
    return (logits_to_probabilities_multilabel(logits) > 0.5).astype(dtype)


def cross_entropy_multiclass(targets: np.ndarray, logits: np.ndarray, epsilon=1e-12) -> NumericType:
    probabilities = logits_to_probabilities_multiclass(logits)

    probabilities_clipped = np.clip(probabilities, epsilon, 1. - epsilon)

    nonzero = probabilities_clipped[np.arange(targets.size, dtype=np.int32), targets]

    score = -mean(np.log(nonzero))

    return score


def cross_entropy_multilabel(targets: np.ndarray, logits: np.ndarray, epsilon=1e-12) -> NumericType:
    probabilities = logits_to_probabilities_multilabel(logits)

    probabilities_clipped = np.clip(probabilities, epsilon, 1. - epsilon)

    scores = -targets * np.log(probabilities_clipped) - (1 - targets) * np.log(1 - probabilities_clipped)

    score = -np.mean(np.sum(scores, axis=-1))

    return score


def mean_absolute_error(targets: np.ndarray, predictions: np.ndarray) -> NumericType:
    return float(np.mean(np.abs(targets - predictions)))


def mean_squared_error(targets: np.ndarray, predictions: np.ndarray) -> NumericType:
    return float(np.mean(np.square(targets - predictions)))


def pearson_correlation(targets: np.ndarray, predictions: np.ndarray) -> NumericType:
    correlation, p_value = sp_stats.pearsonr(targets, predictions)

    return correlation


def accuracy(targets: np.ndarray, predictions: np.ndarray) -> NumericType:
    return sk_metrics.accuracy_score(targets, predictions)


def precision(targets: np.ndarray, predictions: np.ndarray, average: Optional[str] = 'weighted') -> NumericType:
    return sk_metrics.precision_score(targets, predictions, average=average, zero_division=0)


def recall(targets: np.ndarray, predictions: np.ndarray, average: Optional[str] = 'weighted') -> NumericType:
    return sk_metrics.recall_score(targets, predictions, average=average, zero_division=0)


def f1_score(targets: np.ndarray, predictions: np.ndarray, average: Optional[str] = 'weighted') -> NumericType:
    return sk_metrics.f1_score(targets, predictions, average=average, zero_division=0)


def jaccard_score(targets: np.ndarray, predictions: np.ndarray, average: Optional[str] = 'weighted') -> NumericType:
    return sk_metrics.jaccard_score(targets, predictions, average=average)


def jaccard_score_multilabel(targets: np.ndarray, predictions: np.ndarray) -> NumericType:
    overlap = np.sum(targets * predictions, axis=1)

    union = np.sum(targets + predictions > 0, axis=1)

    score = float(np.mean(np.divide(overlap, union, out=np.zeros_like(union, dtype=np.float32), where=union != 0)))

    return score


def mean(array: np.ndarray) -> NumericType:
    return float(np.mean(array))


def calculate_thresholds(targets: np.ndarray, probabilities: np.ndarray) -> np.array:
    num_classes = targets.shape[1]

    best_thresholds = np.empty(shape=num_classes, dtype=np.float32)

    for index in range(num_classes):
        precisions, recalls, thresholds = sk_metrics.precision_recall_curve(
            y_true=targets[:, index],
            probas_pred=probabilities[:, index],
        )

        f1_scores_denominator = precisions + recalls
        f1_scores = np.divide(
            2 * precisions * recalls,
            f1_scores_denominator,
            out=np.zeros_like(f1_scores_denominator),
            where=f1_scores_denominator != 0,
        )

        best_thresholds[index] = thresholds[np.nanargmax(f1_scores) - 1]

    return best_thresholds


def calculate_with_optimized_predictions(
        targets: np.ndarray,
        probabilities: np.ndarray,
        function: AggregatorType,
) -> NumericType:

    thresholds = calculate_thresholds(targets, probabilities)

    predictions = probabilities >= thresholds

    score = function(targets, predictions.astype(np.float32))

    return score
